package net.knarcraft.bookswithoutborders;

import net.knarcraft.bookswithoutborders.command.CommandBooksWithoutBorders;
import net.knarcraft.bookswithoutborders.command.CommandCopy;
import net.knarcraft.bookswithoutborders.command.CommandDecrypt;
import net.knarcraft.bookswithoutborders.command.CommandDelete;
import net.knarcraft.bookswithoutborders.command.CommandDeletePublic;
import net.knarcraft.bookswithoutborders.command.CommandEncrypt;
import net.knarcraft.bookswithoutborders.command.CommandGive;
import net.knarcraft.bookswithoutborders.command.CommandGivePublic;
import net.knarcraft.bookswithoutborders.command.CommandGroupEncrypt;
import net.knarcraft.bookswithoutborders.command.CommandLoad;
import net.knarcraft.bookswithoutborders.command.CommandLoadPublic;
import net.knarcraft.bookswithoutborders.command.CommandReload;
import net.knarcraft.bookswithoutborders.command.CommandSave;
import net.knarcraft.bookswithoutborders.command.CommandSavePublic;
import net.knarcraft.bookswithoutborders.command.CommandSetAuthor;
import net.knarcraft.bookswithoutborders.command.CommandSetBookPrice;
import net.knarcraft.bookswithoutborders.command.CommandSetLore;
import net.knarcraft.bookswithoutborders.command.CommandSetTitle;
import net.knarcraft.bookswithoutborders.command.CommandUnSign;
import net.knarcraft.bookswithoutborders.command.GiveTabCompleter;
import net.knarcraft.bookswithoutborders.listener.PlayerEventListener;
import net.knarcraft.bookswithoutborders.listener.SignEventListener;
import net.knarcraft.bookswithoutborders.state.BookDirectory;
import net.knarcraft.bookswithoutborders.utility.BookToFromTextHelper;
import net.knarcraft.bookswithoutborders.utility.EconomyHelper;
import net.knarcraft.bookswithoutborders.utility.FileHelper;
import org.bukkit.Material;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.command.ConsoleCommandSender;
import org.bukkit.command.PluginCommand;
import org.bukkit.command.TabCompleter;
import org.bukkit.configuration.Configuration;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemFactory;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.BookMeta;
import org.bukkit.plugin.PluginManager;
import org.bukkit.plugin.java.JavaPlugin;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getBookFolder;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getErrorColor;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getSlash;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getSuccessColor;
import static net.knarcraft.bookswithoutborders.utility.InputCleaningHelper.cleanString;

public class BooksWithoutBorders extends JavaPlugin {

    private static int bookDuplicateLimit;
    private static String titleAuthorSeparator;
    private static String loreSeparator;
    private static List<String> firstBooks;
    private static String welcomeMessage;
    private static Material bookPriceType = null;
    private static double bookPriceQuantity;
    private static boolean authorOnlyCopy;
    private static boolean useYml;
    private static boolean adminDecrypt;

    private static ItemFactory itemFactory;
    private static Map<String, List<String>> playerBooksList;
    private static List<String> publicBooksList;
    private static BooksWithoutBorders booksWithoutBorders;
    private static ConsoleCommandSender consoleSender;

    /**
     * Gets the console sender for printing to the console
     * @return <p>The console's console sender</p>
     */
    public static ConsoleCommandSender getConsoleSender() {
        return consoleSender;
    }

    /**
     * Gets available books
     * @param sender <p>The sender wanting to see available books</p>
     * @param getPublic <p>Whether to get available public books</p>
     * @return <p>A list of available books</p>
     */
    public static List<String> getAvailableBooks(CommandSender sender, boolean getPublic) {
        if (getPublic) {
            return new ArrayList<>(publicBooksList);
        } else {
            String senderName = sender.getName();
            if (!playerBooksList.containsKey(senderName)) {
                List<String> newFiles = FileHelper.listFiles(sender, false);
                playerBooksList.put(senderName, newFiles);
            }
            return playerBooksList.get(senderName);
        }
    }

    /**
     * Updates available books
     * @param sender <p>The sender to update books for</p>
     * @param updatePublic <p>Whether to update public books</p>
     */
    public static void updateBooks(CommandSender sender, boolean updatePublic) {
        List<String> newFiles = FileHelper.listFiles(sender, updatePublic);
        if (updatePublic) {
            publicBooksList = newFiles;
        } else {
            playerBooksList.put(sender.getName(), newFiles);
        }
    }

    @Override
    public void onEnable() {
        booksWithoutBorders = this;
        consoleSender = this.getServer().getConsoleSender();
        playerBooksList = new HashMap<>();
        publicBooksList = FileHelper.listFiles(consoleSender, true);
        firstBooks = new ArrayList<>();
        BooksWithoutBordersSettings.initialize(this);

        PluginManager pluginManager = this.getServer().getPluginManager();

        if (getSlash() != null && initialize()) {
            pluginManager.registerEvents(new PlayerEventListener(), this);
            pluginManager.registerEvents(new SignEventListener(), this);
        } else {
            this.getPluginLoader().disablePlugin(this);
        }

        registerCommands();
    }

    /**
     * Gets whether only the author of a book should be able to copy it
     * @return <p>Whether only the book author can copy it</p>
     */
    public static boolean getAuthorOnlyCopy() {
        return authorOnlyCopy;
    }

    /**
     * Gets whether to use YML, not TXT, for saving books
     * @return <p>Whether to use YML for saving books</p>
     */
    public static boolean getUseYml() {
        return useYml;
    }

    /**
     * Gets whether admins should be able to decrypt books without a password, and decrypt all group encrypted books
     * @return <p>Whether admins can bypass the encryption password</p>
     */
    public static boolean getAdminDecrypt() {
        return adminDecrypt;
    }

    /**
     * Sets the quantity of items/currency necessary for copying books
     * @param newQuantity <p>The new quantity necessary for payment</p>
     */
    public static void setBookPriceQuantity(double newQuantity) {
        bookPriceQuantity = newQuantity;
    }

    /**
     * Gets the quantity of items/currency necessary for copying books
     * @return <p>The quantity necessary for payment</p>
     */
    public static double getBookPriceQuantity() {
        return bookPriceQuantity;
    }

    /**
     * Sets the item type used for book pricing
     *
     * <p>This item is the one a player has to pay for copying books. AIR is used to denote economy. null is used if
     * payment is disabled. Otherwise, any item can be used.</p>
     *
     * @param newType <p>The new item type to use for book pricing</p>
     */
    public static void setBookPriceType(Material newType) {
        bookPriceType = newType;
    }

    /**
     * Gets the item type used for book pricing
     *
     * <p>This item is the one a player has to pay for copying books. AIR is used to denote economy. null is used if
     * payment is disabled. Otherwise, any item can be used.</p>
     *
     * @return <p>The item type used for book pricing</p>
     */
    public static Material getBookPriceType() {
        return bookPriceType;
    }

    /**
     * Gets the welcome message to show to new players
     * @return <p>The welcome message to show new players</p>
     */
    public static String getWelcomeMessage() {
        return welcomeMessage;
    }

    /**
     * Gets the limit of duplicates for each book
     * @return <p>The book duplicate limit</p>
     */
    public static int getBookDuplicateLimit() {
        return bookDuplicateLimit;
    }

    /**
     * Gets the separator used to split book title from book author
     * @return <p>The separator between title and author</p>
     */
    public static String getTitleAuthorSeparator() {
        return titleAuthorSeparator;
    }

    /**
     * Gets the separator used to denote a newline in a lore string
     * @return <p>The separator used to denote lore newline</p>
     */
    public static String getLoreSeparator() {
        return loreSeparator;
    }

    /**
     * Gets a copy of the list of books to give new players
     * @return <p>The books to give new players</p>
     */
    public static List<String> getFirstBooks() {
        return new ArrayList<>(firstBooks);
    }

    /**
     * Gets an instance of this plugin
     * @return <p>An instance of this plugin</p>
     */
    public static BooksWithoutBorders getInstance() {
        return booksWithoutBorders;
    }

    /**
     * Registers all commands used by this plugin
     */
    private void registerCommands() {
        registerCommand("giveBook", new CommandGive(), new GiveTabCompleter());
        registerCommand("decryptBook", new CommandDecrypt(), null);
        registerCommand("groupEncryptBook", new CommandGroupEncrypt(), null);
        registerCommand("deleteBook", new CommandDelete(), null);
        registerCommand("deletePublicBook", new CommandDeletePublic(), null);
        registerCommand("copyBook", new CommandCopy(), null);
        registerCommand("unSignBook", new CommandUnSign(), null);
        registerCommand("encryptBook", new CommandEncrypt(), null);
        registerCommand("setBookPrice", new CommandSetBookPrice(), null);
        registerCommand("setLore", new CommandSetLore(), null);
        registerCommand("savePublicBook", new CommandSavePublic(), null);
        registerCommand("saveBook", new CommandSave(), null);
        registerCommand("setBookAuthor", new CommandSetAuthor(), null);
        registerCommand("setTitle", new CommandSetTitle(), null);
        registerCommand("loadBook", new CommandLoad(), null);
        registerCommand("loadPublicBook", new CommandLoadPublic(), null);
        registerCommand("booksWithoutBorders", new CommandBooksWithoutBorders(), null);
        registerCommand("reload", new CommandReload(), null);
        registerCommand("givePublicBook", new CommandGivePublic(), null);
    }

    /**
     * Registers a command
     *
     * @param commandName <p>The name of the command to register</p>
     * @param executor    <p>The executor to register for the command</p>
     */
    private void registerCommand(String commandName, CommandExecutor executor, TabCompleter tabCompleter) {
        PluginCommand pluginCommand = this.getCommand(commandName);
        if (pluginCommand != null) {
            pluginCommand.setExecutor(executor);
            if (tabCompleter != null) {
                pluginCommand.setTabCompleter(tabCompleter);
            }
        } else {
            sendErrorMessage(consoleSender, "Failed to register command " + commandName);
        }
    }

    private boolean initialize() {
        //initializes Item Factory
        try {
            itemFactory = this.getServer().getItemFactory();
        } catch (java.lang.NoSuchMethodError nsmE) {
            sendErrorMessage(consoleSender, "Warning! [BooksWithoutBorders] failed to initialize!");
            sendErrorMessage(consoleSender, "Please confirm the correct version of [BooksWithoutBorders] is");
            sendErrorMessage(consoleSender, "being run for this version of bukkit!");
            return false;
        }

        //Pulls config values
        if (!loadConfig()) {
            return false;
        }

        //Save config with loaded values to fix invalid config values
        saveConfigValues();

        return testFileSaving();
    }

    /**
     * Makes sure necessary folders exist
     * @return <p>True if necessary folders exist</p>
     */
    private boolean testFileSaving() {
        File fileTest = new File(getBookFolder());
        File encryptedFileTest = new File(getBookFolder() + "Encrypted" + getSlash());
        if (!fileTest.exists()) {
            try {
                if (!fileTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Saving failed! Aborting...");
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }
        if (!encryptedFileTest.exists()) {
            try {
                if (!encryptedFileTest.mkdir()) {
                    sendErrorMessage(consoleSender, "Saving failed! Aborting...");
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        }
        return true;
    }

    public void saveConfigValues() {
        Configuration config = this.getConfig();
        config.set("Options.Save_Books_in_Yaml_Format", useYml);
        config.set("Options.Max_Number_of_Duplicates", bookDuplicateLimit);
        config.set("Options.Title-Author_Separator", titleAuthorSeparator);
        config.set("Options.Lore_line_separator", loreSeparator);
        config.set("Options.Books_for_new_players", firstBooks);
        config.set("Options.Message_for_new_players", welcomeMessage);
        if (bookPriceType != null) {
            if (bookPriceType != Material.AIR) {
                config.set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
            } else {
                config.set("Options.Price_to_create_book.Item_type", "Economy");
            }
        } else {
            config.set("Options.Price_to_create_book.Item_type", "Item type name");
        }
        config.set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
        config.set("Options.Admin_Auto_Decrypt", adminDecrypt);
        config.set("Options.Author_Only_Copy", authorOnlyCopy);

        //Handles old book and quill settings
        if (config.contains("Options.Require_book_and_quill_to_create_book")) {
            sendSuccessMessage(consoleSender, "[BooksWithoutBorders] Found old config setting \"Require_book_and_quill_to_create_book\"");
            sendSuccessMessage(consoleSender, "Updating to \"Price_to_create_book\" settings");

            if (config.getBoolean("Options.Require_book_and_quill_to_create_book")) {
                bookPriceType = Material.WRITABLE_BOOK;
                bookPriceQuantity = 1;
                config.set("Options.Price_to_create_book.Item_type", bookPriceType.toString());
                config.set("Options.Price_to_create_book.Required_quantity", bookPriceQuantity);
            }

            config.set("Options.Require_book_and_quill_to_create_book", null);
        }
        this.saveConfig();
    }

    public boolean loadConfig() {
        this.reloadConfig();
        Configuration config = this.getConfig();
        try {
            useYml = config.getBoolean("Options.Save_Books_in_Yaml_Format", true);
            bookDuplicateLimit = config.getInt("Options.Max_Number_of_Duplicates", 5);
            titleAuthorSeparator = config.getString("Options.Title-Author_Separator", ",");
            loreSeparator = config.getString("Options.Lore_line_separator", "~");
            adminDecrypt = config.getBoolean("Options.Admin_Auto_Decrypt", false);
            authorOnlyCopy = config.getBoolean("Options.Author_Only_Copy", false);

            //Set books to give new players
            firstBooks = config.getStringList("Options.Books_for_new_players");
            if (config.contains("Options.Book_for_new_players")) {
                firstBooks.add(config.getString("Options.Book_for_new_players"));
            }
            if (firstBooks.isEmpty()) {
                firstBooks.add(" ");
            }

            welcomeMessage = config.getString("Options.Message_for_new_players", " ");

            //Converts string into material
            String sMaterial = config.getString("Options.Price_to_create_book.Item_type", " ");
            if (sMaterial.equalsIgnoreCase("Economy")) {
                if (EconomyHelper.setupEconomy()) {
                    bookPriceType = Material.AIR;
                } else {
                    sendErrorMessage(consoleSender, "BooksWithoutBorders failed to hook into Vault! Book price not set!");
                    bookPriceType = null;
                }
            } else if (!sMaterial.equalsIgnoreCase(" ")) {
                for (Material m : Material.values()) {
                    if (m.toString().equalsIgnoreCase(sMaterial))
                        bookPriceType = m;
                }
            }
            bookPriceQuantity = config.getDouble("Options.Price_to_create_book.Required_quantity", 0);

            //makes sure titleAuthorSeparator is a valid value
            titleAuthorSeparator = cleanString(titleAuthorSeparator);
            if (titleAuthorSeparator.length() != 1) {
                sendErrorMessage(consoleSender, "Title-Author_Separator is set to an invalid value!");
                sendErrorMessage(consoleSender, "Reverting to default value of \",\"");
                titleAuthorSeparator = ",";
                config.set("Options.Title-Author_Separator", titleAuthorSeparator);
            }
        } catch (Exception e) {
            sendErrorMessage(consoleSender, "Warning! Config.yml failed to load!");
            sendErrorMessage(consoleSender, "Try Looking for settings that are missing values!");
            return false;
        }

        return true;
    }

    /**
     * Loads the given book
     * @param sender <p>The command sender trying to load the book</p>
     * @param fileName <p>The index or file name of the book to load</p>
     * @param isSigned <p>Whether to load the book as signed, and not unsigned</p>
     * @param directory <p>The directory to save the book in</p>
     * @return <p>The loaded book</p>
     */
    public ItemStack loadBook(CommandSender sender, String fileName, String isSigned, String directory) {
        return loadBook(sender, fileName, isSigned, directory, 1);
    }

    /**
     * Loads the given book
     * @param sender <p>The command sender trying to load the book</p>
     * @param fileName <p>The index or file name of the book to load</p>
     * @param isSigned <p>Whether to load the book as signed, and not unsigned</p>
     * @param directory <p>The directory to save the book in</p>
     * @param numCopies <p>The number of copies to load</p>
     * @return <p>The loaded book</p>
     */
    public ItemStack loadBook(CommandSender sender, String fileName, String isSigned, String directory, int numCopies) {
        BookDirectory bookDirectory = BookDirectory.getFromString(directory);

        //Find the filename if a book index is given
        try {
            int bookIndex = Integer.parseInt(fileName);
            List<String> availableFiles = getAvailableBooks(sender, bookDirectory == BookDirectory.PUBLIC);
            if (bookIndex <= availableFiles.size()) {
                fileName = availableFiles.get(Integer.parseInt(fileName) - 1);
            }
        } catch (NumberFormatException ignored) {}

        //Get the full path of the book to load
        File file = null;
        if (bookDirectory == BookDirectory.PUBLIC) {
            file = FileHelper.getBookFile(getBookFolder() + fileName);
        } else if (bookDirectory == BookDirectory.PLAYER) {
            file = FileHelper.getBookFile(getBookFolder() + cleanString(sender.getName()) + getSlash() + fileName);
        } else if (bookDirectory == BookDirectory.ENCRYPTED) {
            file = FileHelper.getBookFile(getBookFolder() + "Encrypted" + getSlash() + directory + getSlash() + fileName);
        }
        if (file == null || !file.isFile()) {
            sendErrorMessage(sender, "Incorrect file name!");
            return null;
        }

        //Make sure the player can pay for the book
        if (booksHavePrice() && !sender.hasPermission("bookswithoutborders.bypassBookPrice") &&
                (bookDirectory == BookDirectory.PUBLIC || bookDirectory == BookDirectory.PLAYER) &&
                cannotPayForBookPrinting((Player) sender, numCopies)) {
            return null;
        }

        //Generate a new empty book
        ItemStack book;
        BookMeta bookMetadata = (BookMeta) itemFactory.getItemMeta(Material.WRITTEN_BOOK);
        if (isSigned.equalsIgnoreCase("true")) {
            book = new ItemStack(Material.WRITTEN_BOOK);
        } else {
            book = new ItemStack(Material.WRITABLE_BOOK);
        }

        //Load the book from the given file
        BookToFromTextHelper.bookFromFile(file, bookMetadata);
        if (bookMetadata == null) {
            sendErrorMessage(sender, "File was blank!!");
            return null;
        }

        //Remove "encrypted" from the book lore
        if (bookDirectory == BookDirectory.ENCRYPTED && bookMetadata.hasLore()) {
            List<String> oldLore = bookMetadata.getLore();
            if (oldLore != null) {
                List<String> newLore = new ArrayList<>(oldLore);
                newLore.remove(0);
                bookMetadata.setLore(newLore);
            }
        }

        //Set the metadata and amount to the new book
        book.setItemMeta(bookMetadata);
        book.setAmount(numCopies);

        return book;
    }

    /**
     * Makes the player pay for printing a given number of books
     *
     * @param player    <p>The player printing the books</p>
     * @param numCopies <p>The number of copies the player is trying to print</p>
     * @return <p>True if the player cannot pay for the printing of the books</p>
     */
    public boolean cannotPayForBookPrinting(Player player, int numCopies) {
        //BookPriceQuantity: How many items are required to pay for each book
        //BookPriceType: Which item is used to pay for the books. AIR = use economy
        double cost = BooksWithoutBorders.bookPriceQuantity * numCopies;
        int itemCost = (int) cost;

        if (BooksWithoutBorders.bookPriceType == Material.AIR) {
            return !EconomyHelper.payForBookPrintingEconomy(player, cost, numCopies);
        } else {
            if (player.getInventory().contains(BooksWithoutBorders.bookPriceType, itemCost)) {
                payForBookPrintingItem(player, itemCost);
                return false;
            } else {
                BooksWithoutBorders.sendErrorMessage(player, itemCost + " " + BooksWithoutBorders.bookPriceType.toString() +
                        "(s) are required for this command!");
                return true;
            }
        }
    }

    /**
     * Takes payment for printing a number of books by withdrawing the correct item
     *
     * @param player   <p>The player which needs to pay</p>
     * @param itemCost <p>The number of items to pay</p>
     */
    private void payForBookPrintingItem(Player player, int itemCost) {
        PlayerInventory playerInventory = player.getInventory();

        int clearedAmount = 0;
        while (clearedAmount < itemCost) {
            int firstItemIndex = playerInventory.first(BooksWithoutBorders.bookPriceType);
            ItemStack firstItem = playerInventory.getItem(firstItemIndex);

            if (Objects.requireNonNull(firstItem).getAmount() <= itemCost - clearedAmount) {
                clearedAmount += firstItem.getAmount();
                player.getInventory().clear(firstItemIndex);
            } else {
                clearedAmount = itemCost;
                firstItem.setAmount(firstItem.getAmount() - (clearedAmount));
            }
        }
    }

    /**
     * Sends a success message to a command sender (player or a console)
     *
     * @param sender  <p>The sender to send the message to</p>
     * @param message <p>The message to send</p>
     */
    public static void sendSuccessMessage(CommandSender sender, String message) {
        sender.sendMessage(getSuccessColor() + message);
    }

    /**
     * Sends an error message to a command sender (player or a console)
     *
     * @param sender  <p>The sender to send the message to</p>
     * @param message <p>The message to send</p>
     */
    public static void sendErrorMessage(CommandSender sender, String message) {
        sender.sendMessage(getErrorColor() + message);
    }

    /**
     * Checks whether books have a price for printing them
     *
     * @return <p>True if players need to pay for printing books</p>
     */
    public boolean booksHavePrice() {
        return (bookPriceType != null && bookPriceQuantity > 0);
    }

}
