package net.knarcraft.bookswithoutborders.command;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.state.EncryptionStyle;
import net.knarcraft.bookswithoutborders.utility.EncryptionHelper;
import net.knarcraft.bookswithoutborders.utility.InventoryHelper;
import org.bukkit.command.Command;
import org.bukkit.command.CommandExecutor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.BookMeta;

/**
 * Command executor for the encrypt command
 */
public class CommandEncrypt implements CommandExecutor {

    @Override
    public boolean onCommand(CommandSender sender, Command command, String label, String[] args) {
        if (performPreChecks(sender, args, 1, "You must specify a key to encrypt a book!") == null) {
            return false;
        }

        EncryptionStyle encryptionStyle = args.length == 2 ? EncryptionStyle.getFromString(args[1]) : EncryptionStyle.SUBSTITUTION;
        return encryptBook(encryptionStyle, (Player) sender, args[0], "");
    }

    /**
     * Performs necessary pre-checks before going through with the encryption
     *
     * @param sender                <p>The sender trying to encrypt a book</p>
     * @param args                  <p>The arguments given</p>
     * @param necessaryArguments    <p>How many arguments is the minimum requirement</p>
     * @param missingArgumentsError <p>The error to show if a required argument is missing</p>
     * @return <p>The metadata of the book to encrypt, or null if any checks fail</p>
     */
    BookMeta performPreChecks(CommandSender sender, String[] args, int necessaryArguments, String missingArgumentsError) {
        if (!(sender instanceof Player player)) {
            BooksWithoutBorders.sendErrorMessage(sender, "This command can only be used by a player!");
            return null;
        }

        if (InventoryHelper.notHoldingOneWrittenBookCheck(player,
                "You must be holding a written book to encrypt it!",
                "You cannot encrypt two books at once!")) {
            return null;
        }

        int argumentCount = args.length;
        if (argumentCount < necessaryArguments) {
            BooksWithoutBorders.sendErrorMessage(player, missingArgumentsError);
            return null;
        }
        if (argumentCount > necessaryArguments + 1) {
            BooksWithoutBorders.sendErrorMessage(player, "Too many command options specified!");
            return null;
        }

        ItemStack heldBook = InventoryHelper.getHeldBook(player, true);
        BookMeta bookMetadata = (BookMeta) heldBook.getItemMeta();
        if (bookMetadata == null) {
            BooksWithoutBorders.sendErrorMessage(player, "Your book seems to be corrupt!");
            return null;
        }
        if (!bookMetadata.hasPages()) {
            BooksWithoutBorders.sendErrorMessage(player, "Book must have contents to encrypt!");
            return null;
        }
        return bookMetadata;
    }

    /**
     * Encrypts the given book
     *
     * @param encryptionStyle <p>The encryption style to use</p>
     * @param player          <p>The player encrypting the book</p>
     * @param key             <p>The encryption key to use</p>
     * @param group           <p>The group to encrypt for</p>
     * @return <p>True if the book was encrypted successfully</p>
     */
    boolean encryptBook(EncryptionStyle encryptionStyle, Player player, String key, String group) {
        ItemStack encryptedBook = EncryptionHelper.encryptBook(player, true, key, encryptionStyle, group);

        if (encryptedBook != null) {
            InventoryHelper.setHeldWrittenBook(player, encryptedBook);
            return true;
        } else {
            return false;
        }
    }

}
