package net.knarcraft.bookswithoutborders.listener;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import net.knarcraft.bookswithoutborders.state.EncryptionStyle;
import net.knarcraft.bookswithoutborders.utility.EncryptionHelper;
import net.knarcraft.bookswithoutborders.utility.FileHelper;
import net.knarcraft.bookswithoutborders.utility.InputCleaningHelper;
import org.bukkit.ChatColor;
import org.bukkit.Material;
import org.bukkit.Tag;
import org.bukkit.block.Sign;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.Action;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.bukkit.inventory.EquipmentSlot;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.BookMeta;

import java.io.File;

import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getBookFolder;
import static net.knarcraft.bookswithoutborders.BooksWithoutBordersSettings.getSlash;
import static net.knarcraft.bookswithoutborders.utility.FileHelper.isBookListIndex;

public class SignEventListener implements Listener {

    private final String slash = getSlash();

    @EventHandler
    public void onSignChange(SignChangeEvent event) {
        if (event.isCancelled()) {
            return;
        }

        String[] lines = event.getLines();
        Player player = event.getPlayer();

        //Check if creating a Books Without Borders Sign and if the player has permission
        if (!lines[0].equalsIgnoreCase("[BwB]") || !player.hasPermission("bookswithoutborders.signs")) {
            return;
        }

        //Mark the sign as active
        event.setLine(0, ChatColor.DARK_GREEN + "[BwB]");

        //Check if the sign is of a valid type
        if (!((lines[1].equalsIgnoreCase("[Encrypt]") || lines[1].equalsIgnoreCase("[Decrypt]") ||
                lines[1].equalsIgnoreCase("[Give]")) && lines[2].trim().length() > 0)) {
            //Mark the second line as invalid
            event.setLine(1, ChatColor.DARK_RED + lines[1]);
            return;
        }

        //Mark the second line as valid
        event.setLine(1, ChatColor.DARK_BLUE + lines[1]);
        lines = event.getLines();

        //Mark valid encryption/decryption sign
        if (lines[1].equalsIgnoreCase(ChatColor.DARK_BLUE + "[Encrypt]") ||
                lines[1].equalsIgnoreCase(ChatColor.DARK_BLUE + "[Decrypt]")) {
            event.setLine(2, ChatColor.MAGIC + lines[2]);
            event.setLine(3, ChatColor.DARK_BLUE + lines[3]);
        } else if (lines[1].equalsIgnoreCase(ChatColor.DARK_BLUE + "[Give]")) {
            //Generate book giving sign
            generateGiveSign(event, lines, player);
        }
    }

    @EventHandler
    public void onClick(PlayerInteractEvent event) {
        if (event.getClickedBlock() == null) {
            return;
        }

        Material clickedBlockType = event.getClickedBlock().getType();
        Player player = event.getPlayer();
        PlayerInventory playerInventory = player.getInventory();
        EquipmentSlot hand = event.getHand();
        if (hand == null) {
            return;
        }
        ItemStack heldItem = playerInventory.getItem(hand);
        Material heldItemType = heldItem.getType();

        if (event.getAction() == Action.RIGHT_CLICK_BLOCK && (Tag.SIGNS.isTagged(clickedBlockType) ||
                Tag.WALL_SIGNS.isTagged(clickedBlockType))) {
            //The player right-clicked a sign
            Sign sign = (Sign) event.getClickedBlock().getState();
            if (signLineEquals(sign, 0, "[BwB]", ChatColor.DARK_GREEN)) {
                if (signLineEquals(sign, 1, "[Encrypt]", ChatColor.DARK_BLUE)) {
                    encryptHeldBookUsingSign(sign, heldItemType, player, hand);
                } else if (signLineEquals(sign, 1, "[Decrypt]", ChatColor.DARK_BLUE)) {
                    decryptHeldBookUsingSign(sign, heldItemType, player, hand);
                } else if (signLineEquals(sign, 1, "[Give]", ChatColor.DARK_BLUE) &&
                        getSignLineColor(sign, 2) == ChatColor.DARK_GREEN) {
                    giveBook(sign, player);
                } else {
                    player.sendMessage("Sign command " + sign.getLine(1) + " " + sign.getLine(2) + " invalid");
                    player.sendMessage(String.valueOf(getSignLineColor(sign, 2)));
                }
            }
        } else if (heldItemType == Material.WRITTEN_BOOK && (event.getAction() == Action.LEFT_CLICK_AIR
                || event.getAction() == Action.LEFT_CLICK_BLOCK)) {
            BookMeta oldBook = (BookMeta) heldItem.getItemMeta();
            if (oldBook == null) {
                return;
            }

            decryptBook(oldBook, player, heldItem, hand);
        }
    }

    /**
     * Decrypts and replaces the player's used book
     *
     * @param sign         <p>The clicked sign</p>
     * @param heldItemType <p>The item type used to click the sign</p>
     * @param player       <p>The player which clicked the sign</p>
     * @param hand         <p>The EquipmentSlot of the used hand</p>
     */
    private void decryptHeldBookUsingSign(Sign sign, Material heldItemType, Player player, EquipmentSlot hand) {
        //Decrypt the held book and replace it
        if (heldItemType == Material.WRITTEN_BOOK) {
            player.closeInventory();

            //Converts user supplied key into integer form
            String lineText = ChatColor.stripColor(sign.getLine(2));
            String key = EncryptionHelper.getNumberKeyFromStringKey(lineText);

            ItemStack book = EncryptionHelper.loadEncryptedBook(player, key, false);
            if (book != null) {
                player.getInventory().setItem(hand, book);
                player.sendMessage(ChatColor.GREEN + "Book decrypted!");
            }
        }
    }

    /**
     * Gets the color of a line on a sign
     *
     * @param sign       <p>The sign to check</p>
     * @param lineNumber <p>The line number to check</p>
     * @return <p>The color of the sign</p>
     */
    private ChatColor getSignLineColor(Sign sign, int lineNumber) {
        String line = sign.getLine(lineNumber);
        if (!ChatColor.stripColor(line).equals(line)) {
            return ChatColor.getByChar(sign.getLine(lineNumber).substring(1, 2));
        } else {
            return null;
        }
    }

    /**
     * Checks if a line on a sign equals some string, and that the sign is the correct color
     *
     * @param sign       <p>The sign to read</p>
     * @param lineNumber <p>The sign line to read</p>
     * @param compareTo  <p>The string to look for</p>
     * @param color      <p>The color to match</p>
     * @return <p>True if the given string is what's on the sign</p>
     */
    private boolean signLineEquals(Sign sign, int lineNumber, String compareTo, ChatColor color) {
        String line = sign.getLine(lineNumber);
        return line.equalsIgnoreCase(color + compareTo);
    }

    /**
     * Changes an edited sign into a sign to give books
     *
     * @param event  <p>The event caused by changing the sign</p>
     * @param lines  <p>The lines on the sign</p>
     * @param player <p>The player which edited the sign</p>
     */
    private void generateGiveSign(SignChangeEvent event, String[] lines, Player player) {
        if (lines[2].length() > 13 || lines[3].length() > 13) {
            player.sendMessage(ChatColor.RED + "[Give] signs' 3rd and 4th lines must be 13 characters or less!");
            markGiveSignValidity(event, false);
            return;
        }

        //Tests if a full file name has been supplied and points to an actual file
        String signFile = getBookFolder() + lines[2] + lines[3];
        if (FileHelper.bookFileExists(signFile)) {
            markGiveSignValidity(event, true);
            return;
        } else {
            if (isBookListIndex(lines[2])) {
                BooksWithoutBorders.updateBooks(player, true);
                markGiveSignValidity(event, true);
                return;
            }
        }
        markGiveSignValidity(event, false);
    }

    /**
     * Marks a give sign as valid or invalid
     *
     * @param event   <p>The event causing the creation of the give sign</p>
     * @param isValid <p>Whether the created sign is valid</p>
     */
    private void markGiveSignValidity(SignChangeEvent event, boolean isValid) {
        String[] lines = event.getLines();
        if (isValid) {
            event.setLine(2, ChatColor.DARK_GREEN + lines[2]);
            event.setLine(3, ChatColor.DARK_GREEN + lines[3]);
        } else {
            event.setLine(2, ChatColor.DARK_RED + lines[2]);
            event.setLine(3, ChatColor.DARK_RED + lines[3]);
        }
    }

    /**
     * Decrypts a book decryptable for a given group
     *
     * @param oldBook  <p>Metadata for the encrypted book</p>
     * @param player   <p>The player decrypting the book</p>
     * @param heldItem <p>The type of the held book</p>
     * @param hand     <p>The hand the player is using to hold the book</p>
     */
    private void decryptBook(BookMeta oldBook, Player player, ItemStack heldItem, EquipmentSlot hand) {
        ItemStack newBook;

        //Check if the book is encrypted by Books Without Borders
        if (!oldBook.hasLore() || oldBook.getLore() == null || !oldBook.getLore().get(0).contains(" encrypted]")) {
            return;
        }

        String groupName = oldBook.getLore().get(0).substring(3).split(" encrypted")[0];

        //Permission check
        if (!player.hasPermission("bookswithoutborders.decrypt." + groupName) &&
                !(BooksWithoutBorders.getAdminDecrypt() && player.hasPermission("bookswithoutborders.admin"))) {
            return;
        }

        String fileName = oldBook.getTitle() + BooksWithoutBorders.getTitleAuthorSeparator() + oldBook.getAuthor();

        String encryptionFile = InputCleaningHelper.cleanString(groupName) + slash + fileName + ".yml";

        File file = new File(getBookFolder() + "Encrypted" + slash + encryptionFile);
        if (!file.isFile()) {
            file = new File(getBookFolder() + fileName + ".txt");
            if (!file.isFile()) {
                return;
            }
        }
        newBook = BooksWithoutBorders.getInstance().loadBook(player, fileName, "true", groupName, heldItem.getAmount());

        if (newBook == null) {
            return;
        }

        player.getInventory().setItem(hand, newBook);
        player.closeInventory();
        player.sendMessage(ChatColor.GREEN + "Book auto-decrypted!");
    }

    /**
     * Encrypts and replaces the player's used book
     *
     * @param sign         <p>The clicked sign</p>
     * @param heldItemType <p>The item type used to click the sign</p>
     * @param player       <p>The player which clicked the sign</p>
     * @param hand         <p>The EquipmentSlot of the used hand</p>
     */
    private void encryptHeldBookUsingSign(Sign sign, Material heldItemType, Player player, EquipmentSlot hand) {
        ItemStack eBook;
        String[] lines = sign.getLines();
        boolean mainHand = hand == EquipmentSlot.HAND;
        if (heldItemType == Material.WRITTEN_BOOK) {
            player.closeInventory();
            eBook = EncryptionHelper.encryptBook(player, mainHand, ChatColor.stripColor(lines[2]),
                    EncryptionStyle.getFromString(ChatColor.stripColor(lines[3])));
            if (eBook != null) {
                player.getInventory().setItem(hand, eBook);
            }
        }
    }

    /**
     * Gives a player the book specified on a sign
     *
     * @param sign   <p>The sign the user clicked</p>
     * @param player <p>The player which clicked the sign</p>
     */
    private void giveBook(Sign sign, Player player) {
        String fileName = ChatColor.stripColor(sign.getLine(2));
        boolean isLoadListNumber = false;

        try {
            Integer.parseInt(fileName);
            isLoadListNumber = true;
        } catch (NumberFormatException ignored){}

        //Add the third line to the second line for the full filename
        String thirdLine = sign.getLine(3);
        if (!isLoadListNumber && thirdLine.length() >= 2) {
            fileName += ChatColor.stripColor(thirdLine);
        }

        ItemStack newBook = BooksWithoutBorders.getInstance().loadBook(player, fileName, "true", "public");

        if (newBook != null) {
            player.getInventory().addItem(newBook);
            player.sendMessage(ChatColor.GREEN + "Received book!");
        } else {
            player.sendMessage(ChatColor.RED + "Book failed to load!");
        }
    }

}
