package net.knarcraft.bookswithoutborders.utility;

import net.knarcraft.bookswithoutborders.BooksWithoutBorders;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.inventory.meta.BookMeta;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import static net.knarcraft.bookswithoutborders.utility.InputCleaningHelper.fixName;

/**
 * Helper class for converting books to and from text/yml files
 */
public final class BookToFromTextHelper {

    private BookToFromTextHelper() {
    }

    /**
     * Saves a book's contents to a .yml file
     *
     * @param path         <p>The path of the folder to save to. Must end with a slash</p>
     * @param fileName     <p>The name of the file to load to</p>
     * @param bookMetadata <p>Metadata about the book to save</p>
     * @throws IOException <p>If unable to save the book</p>
     */
    public static void bookToYml(String path, String fileName, BookMeta bookMetadata) throws IOException {
        FileConfiguration bookYml = YamlConfiguration.loadConfiguration(new File(path, "blank"));

        if (bookMetadata.hasTitle()) {
            bookYml.set("Title", bookMetadata.getTitle());
        }
        if (bookMetadata.hasAuthor()) {
            bookYml.set("Author", bookMetadata.getAuthor());
        }
        if (bookMetadata.hasPages()) {
            bookYml.set("Pages", bookMetadata.getPages());
        }
        if (bookMetadata.hasLore()) {
            bookYml.set("Lore", bookMetadata.getLore());
        }

        bookYml.save(path + fileName + ".yml");
    }

    /**
     * Loads a book from a file
     * @param file <p>The file to load</p>
     * @param bookMetadata <p>The book metadata to use for saving the book</p>
     * @return <p>The book metadata of the loaded book</p>
     */
    public static BookMeta bookFromFile(File file, BookMeta bookMetadata) {
        if (file.getName().endsWith(".txt")) {
            return bookFromTXT(file.getName(), file, bookMetadata);
        } else if (file.getName().endsWith(".yml")) {
            return bookFromYml(file, bookMetadata);
        } else {
            throw new IllegalArgumentException("Trying to load a book file with an unrecognized extension");
        }
    }

    /**
     * Saves a book's contents to a text file
     *
     * @param folderPath   <p>The folder path to save to. Must end with a slash</p>
     * @param fileName     <p>The name of the file to save to</p>
     * @param bookMetadata <p>Metadata about the book to save</p>
     * @throws IOException <p>If unable to save the book</p>
     */
    public static void bookToTXT(String folderPath, String fileName, BookMeta bookMetadata) throws IOException {
        FileWriter fileWriter = new FileWriter(folderPath + fileName + ".txt");
        PrintWriter printWriter = new PrintWriter(fileWriter);
        List<String> pages = bookMetadata.getPages();

        //Save each page of the book as a text line
        printWriter.println("[Book]");
        for (String page : pages) {
            printWriter.println(page);
        }
        printWriter.close();
    }

    /**
     * Loads a book from a .yml file
     *
     * @param file         <p>The path of the file to load</p>
     * @param bookMetadata <p>Metadata which will be altered with the book's contents</p>
     * @return <p>Metadata for the loaded book</p>
     */
    private static BookMeta bookFromYml(File file, BookMeta bookMetadata) {
        try {
            FileConfiguration bookYml = YamlConfiguration.loadConfiguration(file);

            bookMetadata.setTitle(bookYml.getString("Title", "Untitled"));
            bookMetadata.setAuthor(bookYml.getString("Author", "Unknown"));
            bookMetadata.setPages(bookYml.getStringList("Pages"));
            bookMetadata.setLore(bookYml.getStringList("Lore"));
        } catch (IllegalArgumentException e) {
            return null;
        }
        return bookMetadata;
    }

    /**
     * Loads a book from a text file
     *
     * @param fileName     <p>The name of the file to load. Used to create author and title</p>
     * @param file         <p>The file to load</p>
     * @param bookMetadata <p>Metadata which will be altered with the book's contents</p>
     * @return <p>Metadata for the loaded book</p>
     */
    private static BookMeta bookFromTXT(String fileName, File file, BookMeta bookMetadata) {
        String author;
        String title;
        String titleAuthorSeparator = BooksWithoutBorders.getTitleAuthorSeparator();

        //Get title and author from the file name
        if (fileName.contains(titleAuthorSeparator)) {
            author = fileName.substring(fileName.indexOf(titleAuthorSeparator) + 1, fileName.length() - 4);
            title = fileName.substring(0, fileName.indexOf(titleAuthorSeparator));
        } else {
            author = "Unknown";
            title = fileName.substring(0, fileName.length() - 4);
        }

        //Replace underscores with spaces
        title = fixName(title, true);

        //Read the .txt file
        List<String> rawPages;
        try {
            rawPages = readTextFile(file);
        } catch (IOException e) {
            e.printStackTrace();
            return null;
        }

        //Remove any empty pages
        List<String> pages = new ArrayList<>(InputCleaningHelper.cleanList(rawPages));

        //Update the metadata of the book with its new values
        bookMetadata.setAuthor(author);
        bookMetadata.setTitle(title);
        bookMetadata.setPages(pages);

        return bookMetadata;
    }

    /**
     * Reads text from a .txt file
     * @param file <p>The file to read</p>
     * @return <p>A string list where each string is the text on one page</p>
     * @throws IOException <p>If unable to read from the file</p>
     */
    private static List<String> readTextFile(File file) throws IOException {
        List<String> rawPages = new ArrayList<>();
        BufferedReader bufferedReader = new BufferedReader(new FileReader(file));

        //Use the first line to decide if we are loading a book, or creating a new book
        String firstLine = bufferedReader.readLine();
        if (firstLine == null) {
            bufferedReader.close();
            return null;
        }
        if (firstLine.equalsIgnoreCase("[Book]")) {
            //Read every line directly as a page, as this is a saved book
            String readLine;
            do {
                readLine = bufferedReader.readLine();
                if (readLine != null) {
                    rawPages.add(readLine);
                }
            } while (readLine != null);
        } else {
            //Adjust content to page length for each added line to make the text fit the book
            rawPages.add(firstLine);
            while (rawPages.get(rawPages.size() - 1) != null) {
                BookFormatter.formatLastPage(rawPages);
                rawPages.add(bufferedReader.readLine());
            }
            BookFormatter.formatLastPage(rawPages);
        }
        bufferedReader.close();

        return rawPages;
    }

}
